@@ -1,9 +1,7 @@
 //! A graph representation for computing tree diffs.
 
-use std::{
-    cmp::min,
-    hash::{Hash, Hasher},
-};
+use rustc_hash::FxHasher;
+use std::hash::{Hash, Hasher};
 use strsim::normalized_levenshtein;
 
 use crate::syntax::{AtomKind, ChangeKind, Syntax};
@@ -41,11 +39,18 @@ use Edge::*;
 pub struct Vertex<'a> {
     pub lhs_syntax: Option<&'a Syntax<'a>>,
     pub rhs_syntax: Option<&'a Syntax<'a>>,
+    // Instead, store parent pointers for both sides and a hashset of
+    // novel delimiter parent IDs.
+    parents: im_rc::Vector<(Option<&'a Syntax<'a>>, Option<&'a Syntax<'a>>)>,
+    parents_hash: u64,
 }
+
 impl<'a> PartialEq for Vertex<'a> {
     fn eq(&self, other: &Self) -> bool {
         self.lhs_syntax.map(|node| node.id()) == other.lhs_syntax.map(|node| node.id())
             && self.rhs_syntax.map(|node| node.id()) == other.rhs_syntax.map(|node| node.id())
+            && self.parents_hash == other.parents_hash
+            && self.eq_parents_each_side(other)
     }
 }
 impl<'a> Eq for Vertex<'a> {}
@@ -54,12 +59,68 @@ impl<'a> Hash for Vertex<'a> {
     fn hash<H: Hasher>(&self, state: &mut H) {
         self.lhs_syntax.map(|node| node.id()).hash(state);
         self.rhs_syntax.map(|node| node.id()).hash(state);
+
+        self.parents_hash.hash(state);
     }
 }
 
+fn hash_parents(parents: &im_rc::Vector<(Option<&Syntax>, Option<&Syntax>)>) -> u64 {
+    let mut hasher = FxHasher::default();
+
+    for (parent_id, _) in parents.iter() {
+        if let Some(parent_id) = parent_id {
+            // FxHasher finishes with 0 if called with
+            // .write_u32(0). Ensure the u32 written is always
+            // non-zero.
+            hasher.write_u32(parent_id.id() + 1);
+        }
+    }
+    for (_, parent_id) in parents.iter() {
+        if let Some(parent_id) = parent_id {
+            // FxHasher finishes with 0 if called with
+            // .write_u32(0). Ensure the u32 written is always
+            // non-zero.
+            hasher.write_u32(parent_id.id() + 1);
+        }
+    }
+
+    hasher.finish()
+}
+
 impl<'a> Vertex<'a> {
     pub fn is_end(&self) -> bool {
-        self.lhs_syntax.is_none() && self.rhs_syntax.is_none()
+        self.lhs_syntax.is_none() && self.rhs_syntax.is_none() && self.parents.is_empty()
+    }
+
+    fn eq_parents_each_side(&self, other: &Self) -> bool {
+        let self_lhs_parents = self.parents.iter().filter_map(|(lhs, _)| *lhs);
+        let other_lhs_parents = other.parents.iter().filter_map(|(lhs, _)| *lhs);
+        for (self_lhs_parent, other_lhs_parent) in self_lhs_parents.zip(other_lhs_parents) {
+            if self_lhs_parent.id() != other_lhs_parent.id() {
+                return false;
+            }
+        }
+
+        let self_rhs_parents = self.parents.iter().filter_map(|(_, rhs)| *rhs);
+        let other_rhs_parents = other.parents.iter().filter_map(|(_, rhs)| *rhs);
+        for (self_rhs_parent, other_rhs_parent) in self_rhs_parents.zip(other_rhs_parents) {
+            if self_rhs_parent.id() != other_rhs_parent.id() {
+                return false;
+            }
+        }
+
+        true
+    }
+
+    pub fn new(lhs_syntax: Option<&'a Syntax<'a>>, rhs_syntax: Option<&'a Syntax<'a>>) -> Self {
+        let parents = im_rc::Vector::new();
+        let parents_hash = hash_parents(&parents);
+        Vertex {
+            lhs_syntax,
+            rhs_syntax,
+            parents,
+            parents_hash,
+        }
     }
 }
 
@@ -72,24 +133,32 @@ impl<'a> Vertex<'a> {
 /// See [`neighbours`] for all the edges available for a given `Vertex`.
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
 pub enum Edge {
-    UnchangedNode { depth_difference: u32 },
-    UnchangedDelimiter { depth_difference: u32 },
+    UnchangedNode,
+    EnterUnchangedDelimiter,
     ReplacedComment { levenshtein_pct: u8 },
     NovelAtomLHS { contiguous: bool },
     NovelAtomRHS { contiguous: bool },
-    NovelDelimiterLHS { contiguous: bool },
-    NovelDelimiterRHS { contiguous: bool },
+    // TODO: An EnterNovelDelimiterBoth edge might help performance
+    // rather doing LHS and RHS separately.
+    EnterNovelDelimiterLHS { contiguous: bool },
+    EnterNovelDelimiterRHS { contiguous: bool },
     NovelTreeLHS { num_descendants: u32 },
     NovelTreeRHS { num_descendants: u32 },
+    ExitDelimiterLHS,
+    ExitDelimiterRHS,
+    ExitDelimiterBoth,
 }
 
 impl Edge {
     pub fn cost(&self) -> u64 {
         match self {
+            ExitDelimiterBoth => 1,
+            ExitDelimiterLHS | ExitDelimiterRHS => 2,
+
             // Matching nodes is always best.
-            UnchangedNode { depth_difference } => min(40, *depth_difference as u64 + 1),
+            UnchangedNode => 1,
             // Matching an outer delimiter is good.
-            UnchangedDelimiter { depth_difference } => 100 + min(40, *depth_difference as u64),
+            EnterUnchangedDelimiter => 100,
 
             // Replacing a comment is better than treating it as novel.
             ReplacedComment { levenshtein_pct } => 150 + u64::from(100 - levenshtein_pct),
@@ -97,8 +166,8 @@ impl Edge {
             // Otherwise, we've added/removed a node.
             NovelAtomLHS { contiguous }
             | NovelAtomRHS { contiguous }
-            | NovelDelimiterLHS { contiguous }
-            | NovelDelimiterRHS { contiguous } => {
+            | EnterNovelDelimiterLHS { contiguous }
+            | EnterNovelDelimiterRHS { contiguous } => {
                 if *contiguous {
                     300
                 } else {
@@ -131,19 +200,80 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
     }
 
     let mut i = 0;
+
+    if let Some((lhs_parent, rhs_parent)) = v.parents.back() {
+        match (lhs_parent, rhs_parent) {
+            (Some(lhs_parent), Some(rhs_parent))
+                if v.lhs_syntax.is_none() && v.rhs_syntax.is_none() =>
+            {
+                // We have exhausted all the nodes on both lists, so we can
+                // move up to the parent node.
+                let mut next_parents = v.parents.clone();
+                next_parents.pop_back();
+                let parents_hash = hash_parents(&next_parents);
+
+                // Continue from sibling of parent.
+                buf[i] = Some((
+                    ExitDelimiterBoth,
+                    Vertex {
+                        lhs_syntax: lhs_parent.next_sibling(),
+                        rhs_syntax: rhs_parent.next_sibling(),
+                        parents: next_parents,
+                        parents_hash,
+                    },
+                ));
+                i += 1;
+            }
+            (Some(lhs_parent), None) if v.lhs_syntax.is_none() => {
+                // Move to next after LHS parent.
+                let mut next_parents = v.parents.clone();
+                next_parents.pop_back();
+                let parents_hash = hash_parents(&next_parents);
+
+                // Continue from sibling of parent.
+                buf[i] = Some((
+                    ExitDelimiterLHS,
+                    Vertex {
+                        lhs_syntax: lhs_parent.next_sibling(),
+                        rhs_syntax: v.rhs_syntax,
+                        parents: next_parents,
+                        parents_hash,
+                    },
+                ));
+                i += 1;
+            }
+            (None, Some(rhs_parent)) if v.rhs_syntax.is_none() => {
+                // Move to next after RHS parent.
+                let mut next_parents = v.parents.clone();
+                next_parents.pop_back();
+                let parents_hash = hash_parents(&next_parents);
+
+                // Continue from sibling of parent.
+                buf[i] = Some((
+                    ExitDelimiterRHS,
+                    Vertex {
+                        lhs_syntax: v.lhs_syntax,
+                        rhs_syntax: rhs_parent.next_sibling(),
+                        parents: next_parents,
+                        parents_hash,
+                    },
+                ));
+                i += 1;
+            }
+            _ => {}
+        }
+    }
+
     if let (Some(lhs_syntax), Some(rhs_syntax)) = (&v.lhs_syntax, &v.rhs_syntax) {
         if lhs_syntax == rhs_syntax {
-            let depth_difference = (lhs_syntax.num_ancestors() as i32
-                - rhs_syntax.num_ancestors() as i32)
-                .abs() as u32;
-
             // Both nodes are equal, the happy case.
-            // TODO: this is only OK if we've not changed depth.
             buf[i] = Some((
-                UnchangedNode { depth_difference },
+                UnchangedNode,
                 Vertex {
-                    lhs_syntax: lhs_syntax.next(),
-                    rhs_syntax: rhs_syntax.next(),
+                    lhs_syntax: lhs_syntax.next_sibling(),
+                    rhs_syntax: rhs_syntax.next_sibling(),
+                    parents: v.parents.clone(),
+                    parents_hash: v.parents_hash,
                 },
             ));
             return;
@@ -166,26 +296,21 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
         {
             // The list delimiters are equal, but children may not be.
             if lhs_open_content == rhs_open_content && lhs_close_content == rhs_close_content {
-                let lhs_next = if lhs_children.is_empty() {
-                    lhs_syntax.next()
-                } else {
-                    Some(lhs_children[0])
-                };
-                let rhs_next = if rhs_children.is_empty() {
-                    rhs_syntax.next()
-                } else {
-                    Some(rhs_children[0])
-                };
+                let lhs_next = lhs_children.get(0).copied();
+                let rhs_next = rhs_children.get(0).copied();
 
-                let depth_difference = (lhs_syntax.num_ancestors() as i32
-                    - rhs_syntax.num_ancestors() as i32)
-                    .abs() as u32;
+                // TODO: be consistent between parents_next and next_parents.
+                let mut parents_next = v.parents.clone();
+                parents_next.push_back((Some(lhs_syntax), Some(rhs_syntax)));
+                let parents_hash = hash_parents(&parents_next);
 
                 buf[i] = Some((
-                    UnchangedDelimiter { depth_difference },
+                    EnterUnchangedDelimiter,
                     Vertex {
                         lhs_syntax: lhs_next,
                         rhs_syntax: rhs_next,
+                        parents: parents_next,
+                        parents_hash,
                     },
                 ));
                 i += 1;
@@ -213,8 +338,10 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                 buf[i] = Some((
                     ReplacedComment { levenshtein_pct },
                     Vertex {
-                        lhs_syntax: lhs_syntax.next(),
-                        rhs_syntax: rhs_syntax.next(),
+                        lhs_syntax: lhs_syntax.next_sibling(),
+                        rhs_syntax: rhs_syntax.next_sibling(),
+                        parents: v.parents.clone(),
+                        parents_hash: v.parents_hash,
                     },
                 ));
                 i += 1;
@@ -228,11 +355,15 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
             Syntax::Atom { .. } => {
                 buf[i] = Some((
                     NovelAtomLHS {
+                        // TODO: should this apply if prev is a parent
+                        // node rather than a sibling?
                         contiguous: lhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
-                        lhs_syntax: lhs_syntax.next(),
+                        lhs_syntax: lhs_syntax.next_sibling(),
                         rhs_syntax: v.rhs_syntax,
+                        parents: v.parents.clone(),
+                        parents_hash: v.parents_hash,
                     },
                 ));
                 i += 1;
@@ -243,19 +374,21 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                 num_descendants,
                 ..
             } => {
-                let lhs_next = if children.is_empty() {
-                    lhs_syntax.next()
-                } else {
-                    Some(children[0])
-                };
+                let lhs_next = children.get(0).copied();
+
+                let mut parents_next = v.parents.clone();
+                parents_next.push_back((Some(lhs_syntax), None));
+                let parents_hash = hash_parents(&parents_next);
 
                 buf[i] = Some((
-                    NovelDelimiterLHS {
+                    EnterNovelDelimiterLHS {
                         contiguous: lhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
                         lhs_syntax: lhs_next,
                         rhs_syntax: v.rhs_syntax,
+                        parents: parents_next,
+                        parents_hash,
                     },
                 ));
                 i += 1;
@@ -266,8 +399,10 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                             num_descendants: *num_descendants,
                         },
                         Vertex {
-                            lhs_syntax: lhs_syntax.next(),
+                            lhs_syntax: lhs_syntax.next_sibling(),
                             rhs_syntax: v.rhs_syntax,
+                            parents: v.parents.clone(),
+                            parents_hash: v.parents_hash,
                         },
                     ));
                     i += 1;
@@ -286,9 +421,12 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                     },
                     Vertex {
                         lhs_syntax: v.lhs_syntax,
-                        rhs_syntax: rhs_syntax.next(),
+                        rhs_syntax: rhs_syntax.next_sibling(),
+                        parents: v.parents.clone(),
+                        parents_hash: v.parents_hash,
                     },
                 ));
+                i += 1;
             }
             // Step into this partially/fully novel list.
             Syntax::List {
@@ -296,19 +434,21 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                 num_descendants,
                 ..
             } => {
-                let rhs_next = if children.is_empty() {
-                    rhs_syntax.next()
-                } else {
-                    Some(children[0])
-                };
+                let rhs_next = children.get(0).copied();
+
+                let mut parents_next = v.parents.clone();
+                parents_next.push_back((None, Some(rhs_syntax)));
+                let parents_hash = hash_parents(&parents_next);
 
                 buf[i] = Some((
-                    NovelDelimiterRHS {
+                    EnterNovelDelimiterRHS {
                         contiguous: rhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
                         lhs_syntax: v.lhs_syntax,
                         rhs_syntax: rhs_next,
+                        parents: parents_next,
+                        parents_hash,
                     },
                 ));
                 i += 1;
@@ -320,26 +460,36 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         },
                         Vertex {
                             lhs_syntax: v.lhs_syntax,
-                            rhs_syntax: rhs_syntax.next(),
+                            rhs_syntax: rhs_syntax.next_sibling(),
+                            parents: v.parents.clone(),
+                            parents_hash: v.parents_hash,
                         },
                     ));
+                    i += 1;
                 }
             }
         }
     }
+    assert!(
+        i > 0,
+        "Must always find some next steps if node is not the end"
+    );
 }
 
 pub fn mark_route(route: &[(Edge, Vertex)]) {
     for (e, v) in route {
         match e {
-            UnchangedNode { .. } => {
+            ExitDelimiterBoth | ExitDelimiterLHS | ExitDelimiterRHS => {
+                // Nothing to do: we have already marked this node when we entered it.
+            }
+            UnchangedNode => {
                 // No change on this node or its children.
                 let lhs = v.lhs_syntax.unwrap();
                 let rhs = v.rhs_syntax.unwrap();
                 lhs.set_change_deep(ChangeKind::Unchanged(rhs));
                 rhs.set_change_deep(ChangeKind::Unchanged(lhs));
             }
-            UnchangedDelimiter { .. } => {
+            EnterUnchangedDelimiter => {
                 // No change on the outer delimiter, but children may
                 // have changed.
                 let lhs = v.lhs_syntax.unwrap();
@@ -359,11 +509,11 @@ pub fn mark_route(route: &[(Edge, Vertex)]) {
                     rhs.set_change(ChangeKind::Novel);
                 }
             }
-            NovelAtomLHS { .. } | NovelDelimiterLHS { .. } => {
+            NovelAtomLHS { .. } | EnterNovelDelimiterLHS { .. } => {
                 let lhs = v.lhs_syntax.unwrap();
                 lhs.set_change(ChangeKind::Novel);
             }
-            NovelAtomRHS { .. } | NovelDelimiterRHS { .. } => {
+            NovelAtomRHS { .. } | EnterNovelDelimiterRHS { .. } => {
                 let rhs = v.rhs_syntax.unwrap();
                 rhs.set_change(ChangeKind::Novel);
             }
