@@ -44,7 +44,7 @@ use mimalloc::MiMalloc;
 #[global_allocator]
 static GLOBAL: MiMalloc = MiMalloc;
 
-use options::{configure_color, Mode};
+use options::{should_use_color, Mode};
 use sliders::fix_all_sliders;
 use std::{env, path::Path};
 use style::BackgroundColor;
@@ -118,6 +118,7 @@ fn main() {
         }
         Mode::Diff {
             node_limit,
+            byte_limit,
             print_unchanged,
             missing_as_empty,
             background_color,
@@ -128,17 +129,18 @@ fn main() {
             rhs_path,
             ..
         } => {
-            configure_color(color_output);
+            let use_color = should_use_color(color_output);
 
             let lhs_path = Path::new(&lhs_path);
             let rhs_path = Path::new(&rhs_path);
 
             if lhs_path.is_dir() && rhs_path.is_dir() {
                 for diff_result in
-                    diff_directories(lhs_path, rhs_path, missing_as_empty, node_limit)
+                    diff_directories(lhs_path, rhs_path, missing_as_empty, node_limit, byte_limit)
                 {
                     print_diff_result(
                         display_width,
+                        use_color,
                         background_color,
                         print_unchanged,
                         &diff_result,
@@ -151,9 +153,11 @@ fn main() {
                     rhs_path,
                     missing_as_empty,
                     node_limit,
+                    byte_limit,
                 );
                 print_diff_result(
                     display_width,
+                    use_color,
                     background_color,
                     print_unchanged,
                     &diff_result,
@@ -170,9 +174,10 @@ fn diff_file(
     rhs_path: &Path,
     missing_as_empty: bool,
     node_limit: u32,
+    byte_limit: usize,
 ) -> DiffResult {
     let (lhs_bytes, rhs_bytes) = read_files_or_die(lhs_path, rhs_path, missing_as_empty);
-    diff_file_content(display_path, &lhs_bytes, &rhs_bytes, node_limit)
+    diff_file_content(display_path, &lhs_bytes, &rhs_bytes, node_limit, byte_limit)
 }
 
 fn diff_file_content(
@@ -180,6 +185,7 @@ fn diff_file_content(
     lhs_bytes: &[u8],
     rhs_bytes: &[u8],
     node_limit: u32,
+    byte_limit: usize,
 ) -> DiffResult {
     if is_probably_binary(lhs_bytes) || is_probably_binary(rhs_bytes) {
         return DiffResult {
@@ -227,6 +233,15 @@ fn diff_file_content(
     }
 
     let (lang_name, lhs_positions, rhs_positions) = match ts_lang {
+        _ if lhs_bytes.len() > byte_limit || rhs_bytes.len() > byte_limit => {
+            let lhs_positions = line_parser::change_positions(&lhs_src, &rhs_src);
+            let rhs_positions = line_parser::change_positions(&rhs_src, &lhs_src);
+            (
+                Some("Text (exceeded DFT_BYTE_LIMIT)".into()),
+                lhs_positions,
+                rhs_positions,
+            )
+        }
         Some(ts_lang) => {
             let arena = Arena::new();
             let lhs = tsp::parse(&arena, &lhs_src, &ts_lang);
@@ -295,6 +310,7 @@ fn diff_directories(
     rhs_dir: &Path,
     missing_as_empty: bool,
     node_limit: u32,
+    byte_limit: usize,
 ) -> Vec<DiffResult> {
     let mut res = vec![];
     for entry in WalkDir::new(lhs_dir).into_iter().filter_map(Result::ok) {
@@ -314,13 +330,16 @@ fn diff_directories(
             &rhs_path,
             missing_as_empty,
             node_limit,
+            byte_limit,
         ));
     }
     res
 }
 
+// TODO: factor out a DiffOptions struct.
 fn print_diff_result(
     display_width: usize,
+    use_color: bool,
     background: BackgroundColor,
     print_unchanged: bool,
     summary: &DiffResult,
@@ -344,9 +363,11 @@ fn print_diff_result(
                 if print_unchanged {
                     println!(
                         "{}",
-                        style::header(&summary.path, 1, 1, &lang_name, background)
+                        style::header(&summary.path, 1, 1, &lang_name, use_color, background)
                     );
                     if lang_name == "Text" {
+                        // TODO: there are other Text names now, so
+                        // they will hit the second case incorrectly.
                         println!("No changes.\n");
                     } else {
                         println!("No syntactic changes.\n");
@@ -364,12 +385,14 @@ fn print_diff_result(
                     &hunks,
                     &summary.path,
                     &lang_name,
+                    use_color,
                     background,
                 );
             } else {
                 side_by_side::print(
                     &hunks,
                     display_width,
+                    use_color,
                     background,
                     &summary.path,
                     &lang_name,
@@ -385,7 +408,7 @@ fn print_diff_result(
             if print_unchanged || changed {
                 println!(
                     "{}",
-                    style::header(&summary.path, 1, 1, "binary", background)
+                    style::header(&summary.path, 1, 1, "binary", use_color, background)
                 );
                 if changed {
                     println!("Binary contents changed.")
@@ -398,7 +421,7 @@ fn print_diff_result(
             // We're diffing a binary file against a text file.
             println!(
                 "{}",
-                style::header(&summary.path, 1, 1, "binary", background)
+                style::header(&summary.path, 1, 1, "binary", use_color, background)
             );
             println!("Binary contents changed.")
         }
@@ -423,12 +446,18 @@ fn num_nodes(roots: &[&syntax::Syntax]) -> u32 {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::options::DEFAULT_NODE_LIMIT;
+    use crate::options::{DEFAULT_BYTE_LIMIT, DEFAULT_NODE_LIMIT};
 
     #[test]
     fn test_diff_identical_content() {
         let s = "foo";
-        let res = diff_file_content("foo.el", s.as_bytes(), s.as_bytes(), DEFAULT_NODE_LIMIT);
+        let res = diff_file_content(
+            "foo.el",
+            s.as_bytes(),
+            s.as_bytes(),
+            DEFAULT_NODE_LIMIT,
+            DEFAULT_BYTE_LIMIT,
+        );
 
         assert_eq!(res.lhs_positions, vec![]);
         assert_eq!(res.rhs_positions, vec![]);
