@@ -1,11 +1,11 @@
 //! Apply colours and styling to strings.
 
 use crate::{
-    lines::{codepoint_len, substring_by_codepoint, LineNumber},
+    lines::{codepoint_len, LineNumber},
     positions::SingleLineSpan,
     syntax::{AtomKind, MatchKind, MatchedPos, TokenKind},
 };
-use colored::*;
+use owo_colors::{OwoColorize, Style};
 use std::{
     cmp::{max, min},
     collections::HashMap,
@@ -17,35 +17,30 @@ pub enum BackgroundColor {
     Light,
 }
 
-#[derive(Clone, Copy, Debug)]
-pub struct Style {
-    foreground: Option<Color>,
-    background: Option<Color>,
-    bold: bool,
-    dimmed: bool,
-    italic: bool,
+impl BackgroundColor {
+    pub fn is_dark(&self) -> bool {
+        matches!(self, BackgroundColor::Dark)
+    }
 }
 
-impl Style {
-    fn apply(&self, s: &str) -> String {
-        let mut res = if let Some(foreground) = self.foreground {
-            s.color(foreground)
-        } else {
-            s.normal()
-        };
-        if self.bold {
-            res = res.bold();
-        }
-        if self.dimmed {
-            res = res.dimmed();
-        }
-        if self.italic {
-            res = res.italic();
-        }
-        if let Some(background) = self.background {
-            res = res.on_color(background);
-        };
-        res.to_string()
+/// Slice `s` from `start` inclusive to `end` exclusive by codepoint. This is safer than
+/// slicing by bytes, which panics if the byte isn't on a codepoint
+/// boundary.
+fn substring_by_codepoint(s: &str, start: usize, end: usize) -> &str {
+    if start == end {
+        return &s[0..0];
+    }
+
+    assert!(end > start);
+
+    let mut char_idx_iter = s.char_indices();
+    let byte_start = char_idx_iter
+        .nth(start)
+        .expect("Expected a codepoint index inside `s`.")
+        .0;
+    match char_idx_iter.nth(end - start - 1) {
+        Some(byte_end) => &s[byte_start..byte_end.0],
+        None => &s[byte_start..],
     }
 }
 
@@ -78,13 +73,20 @@ fn highlight_missing_style_bug(s: &str) -> String {
 pub fn split_and_apply(
     line: &str,
     max_len: usize,
+    use_color: bool,
     styles: &[(SingleLineSpan, Style)],
 ) -> Vec<String> {
     if styles.is_empty() && !line.is_empty() {
         // Missing styles is a bug, so highlight in purple to make this obvious.
         return split_string(line, max_len)
             .into_iter()
-            .map(|part| highlight_missing_style_bug(&part))
+            .map(|part| {
+                if use_color {
+                    highlight_missing_style_bug(&part)
+                } else {
+                    part
+                }
+            })
             .collect();
     }
 
@@ -118,7 +120,7 @@ pub fn split_and_apply(
                     max(0, span.start_col as isize - prev_length as isize) as usize,
                     min(codepoint_len(&part), span.end_col - prev_length),
                 );
-                res.push_str(&style.apply(span_s));
+                res.push_str(&span_s.style(*style).to_string());
             }
             i = span.end_col;
         }
@@ -135,19 +137,20 @@ pub fn split_and_apply(
     styled_parts
 }
 
-/// Return a copy of `line` with styles applied to all the spans specified.
-/// Dim any parts of the line that have no spans.
+/// Return a copy of `line` with styles applied to all the spans
+/// specified.
 fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
     if styles.is_empty() && !line.is_empty() {
         return highlight_missing_style_bug(line);
     }
 
+    let line_codepoints = codepoint_len(line);
     let mut res = String::with_capacity(line.len());
     let mut i = 0;
     for (span, style) in styles {
         // The remaining spans are beyond the end of this line. This
         // occurs when we truncate the line to fit on the display.
-        if span.start_col >= codepoint_len(line) {
+        if span.start_col >= line_codepoints {
             break;
         }
 
@@ -158,14 +161,14 @@ fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
 
         // Apply style to the substring in this span.
         let span_s =
-            substring_by_codepoint(line, span.start_col, min(codepoint_len(line), span.end_col));
-        res.push_str(&style.apply(span_s));
+            substring_by_codepoint(line, span.start_col, min(line_codepoints, span.end_col));
+        res.push_str(&span_s.style(*style).to_string());
         i = span.end_col;
     }
 
     // Unstyled text after the last span.
-    if i < codepoint_len(line) {
-        let span_s = substring_by_codepoint(line, i, codepoint_len(line));
+    if i < line_codepoints {
+        let span_s = substring_by_codepoint(line, i, line_codepoints);
         res.push_str(span_s);
     }
     res
@@ -186,8 +189,7 @@ fn group_by_line(
     ranges_by_line
 }
 
-/// Apply the `Style`s to the spans specified. Dim any text that
-/// doesn't have any styles applied.
+/// Apply the `Style`s to the spans specified.
 ///
 /// Tolerant against lines in `s` being shorter than the spans.
 fn apply(s: &str, styles: &[(SingleLineSpan, Style)]) -> String {
@@ -202,74 +204,83 @@ fn apply(s: &str, styles: &[(SingleLineSpan, Style)]) -> String {
     res
 }
 
+pub fn novel_style(style: Style, is_lhs: bool, background: BackgroundColor) -> Style {
+    if background.is_dark() {
+        if is_lhs {
+            style.bright_red()
+        } else {
+            style.bright_green()
+        }
+    } else {
+        if is_lhs {
+            style.red()
+        } else {
+            style.green()
+        }
+    }
+}
+
 pub fn color_positions(
     is_lhs: bool,
     background: BackgroundColor,
     positions: &[MatchedPos],
 ) -> Vec<(SingleLineSpan, Style)> {
-    let red = match background {
-        BackgroundColor::Dark => Color::BrightRed,
-        BackgroundColor::Light => Color::Red,
-    };
-    let green = match background {
-        BackgroundColor::Dark => Color::BrightGreen,
-        BackgroundColor::Light => Color::Green,
-    };
-    let novel_color = if is_lhs { red } else { green };
-
     let mut styles = vec![];
     for pos in positions {
-        let line_pos = pos.pos;
-        let style = match pos.kind {
-            MatchKind::UnchangedToken { highlight, .. } => Style {
-                foreground: match highlight {
-                    TokenKind::Atom(AtomKind::String) => Some(match background {
-                        BackgroundColor::Dark => Color::BrightMagenta,
-                        BackgroundColor::Light => Color::Magenta,
-                    }),
-                    TokenKind::Atom(AtomKind::Comment) => Some(match background {
-                        BackgroundColor::Dark => Color::BrightBlue,
-                        BackgroundColor::Light => Color::Blue,
-                    }),
-                    _ => None,
+        let mut style = Style::new();
+        match pos.kind {
+            MatchKind::UnchangedToken { highlight, .. } => match highlight {
+                TokenKind::Atom(atom_kind) => match atom_kind {
+                    AtomKind::String => {
+                        style = if background.is_dark() {
+                            style.bright_magenta()
+                        } else {
+                            style.magenta()
+                        };
+                    }
+                    AtomKind::Comment => {
+                        style = style.italic();
+                        style = if background.is_dark() {
+                            style.bright_blue()
+                        } else {
+                            style.blue()
+                        };
+                    }
+                    AtomKind::Keyword | AtomKind::Type => {
+                        style = style.bold();
+                    }
+                    _ => {}
                 },
-                background: None,
-                bold: match highlight {
-                    TokenKind::Atom(AtomKind::Keyword) => true,
-                    TokenKind::Atom(AtomKind::Type) => true,
-                    _ => false,
-                },
-                dimmed: false,
-                italic: matches!(highlight, TokenKind::Atom(AtomKind::Comment)),
-            },
-            MatchKind::Novel { highlight, .. } => Style {
-                foreground: Some(novel_color),
-                background: None,
-                bold: match highlight {
-                    TokenKind::Delimiter => true,
-                    TokenKind::Atom(AtomKind::Keyword) => true,
-                    TokenKind::Atom(AtomKind::Type) => true,
-                    TokenKind::Atom(_) => false,
-                },
-                dimmed: false,
-                italic: matches!(highlight, TokenKind::Atom(AtomKind::Comment)),
-            },
-            MatchKind::NovelWord { highlight } => Style {
-                foreground: Some(novel_color),
-                background: None,
-                bold: true,
-                dimmed: false,
-                italic: matches!(highlight, TokenKind::Atom(AtomKind::Comment)),
-            },
-            MatchKind::NovelLinePart { highlight, .. } => Style {
-                foreground: Some(novel_color),
-                background: None,
-                bold: false,
-                dimmed: false,
-                italic: matches!(highlight, TokenKind::Atom(AtomKind::Comment)),
+                _ => {}
             },
+            MatchKind::Novel { highlight, .. } => {
+                style = novel_style(style, is_lhs, background);
+                match highlight {
+                    TokenKind::Delimiter
+                    | TokenKind::Atom(AtomKind::Keyword)
+                    | TokenKind::Atom(AtomKind::Type) => {
+                        style = style.bold();
+                    }
+                    _ => {}
+                }
+                if matches!(highlight, TokenKind::Atom(AtomKind::Comment)) {
+                    style = style.italic();
+                }
+            }
+            MatchKind::NovelWord { highlight } => {
+                style = novel_style(style, is_lhs, background).bold();
+                if matches!(highlight, TokenKind::Atom(AtomKind::Comment)) {
+                    style = style.italic();
+                }
+            }
+            MatchKind::NovelLinePart { highlight, .. } => {
+                style = novel_style(style, is_lhs, background);
+                if matches!(highlight, TokenKind::Atom(AtomKind::Comment)) {
+                    style = style.italic();
+                }
+            }
         };
-        styles.push((line_pos, style));
+        styles.push((pos.pos, style));
     }
     styles
 }
@@ -289,18 +300,23 @@ pub fn header(
     hunk_num: usize,
     hunk_total: usize,
     language_name: &str,
+    use_color: bool,
     background: BackgroundColor,
 ) -> String {
+    let file_name_pretty = if use_color {
+        if background.is_dark() {
+            file_name.bright_yellow().to_string()
+        } else {
+            file_name.yellow().to_string()
+        }
+        .bold()
+        .to_string()
+    } else {
+        file_name.to_string()
+    };
     format!(
         "{} --- {}/{} --- {}",
-        match background {
-            BackgroundColor::Dark => file_name.bright_yellow(),
-            BackgroundColor::Light => file_name.yellow(),
-        }
-        .bold(),
-        hunk_num,
-        hunk_total,
-        language_name
+        file_name_pretty, hunk_num, hunk_total, language_name
     )
 }
 
@@ -309,6 +325,16 @@ mod tests {
     use super::*;
     use pretty_assertions::assert_eq;
 
+    #[test]
+    fn test_substring_by_codepoint() {
+        assert_eq!(substring_by_codepoint("abcd", 0, 2), "ab");
+    }
+
+    #[test]
+    fn test_substring_by_codepoint_empty() {
+        assert_eq!(substring_by_codepoint("abcd", 0, 0), "");
+    }
+
     #[test]
     fn split_string_simple() {
         assert_eq!(split_string("fooba", 3), vec!["foo", "ba "]);
