@@ -2,8 +2,6 @@
 
 #![allow(clippy::mutable_key_type)] // Hash for Syntax doesn't use mutable fields.
 
-use lazy_static::lazy_static;
-use regex::Regex;
 use std::{
     cell::Cell,
     collections::HashMap,
@@ -42,9 +40,8 @@ impl<'a> fmt::Debug for ChangeKind<'a> {
 
 /// Fields that are common to both `Syntax::List` and `Syntax::Atom`.
 pub struct SyntaxInfo<'a> {
-    /// The syntax node that occurs after this one, in a depth-first
-    /// tree traversal.
-    pub next: Cell<Option<&'a Syntax<'a>>>,
+    /// The next node with the same parent as this one.
+    pub next_sibling: Cell<Option<&'a Syntax<'a>>>,
     /// The syntax node that occurs before this one, in a depth-first
     /// tree traversal.
     pub prev: Cell<Option<&'a Syntax<'a>>>,
@@ -71,7 +68,7 @@ pub struct SyntaxInfo<'a> {
 impl<'a> SyntaxInfo<'a> {
     pub fn new() -> Self {
         Self {
-            next: Cell::new(None),
+            next_sibling: Cell::new(None),
             prev: Cell::new(None),
             parent: Cell::new(None),
             prev_is_contiguous: Cell::new(false),
@@ -145,12 +142,12 @@ impl<'a> fmt::Debug for Syntax<'a> {
                 if env::var("DFT_VERBOSE").is_ok() {
                     ds.field("change", &info.change.get());
 
-                    let next_s = match info.next.get() {
+                    let next_sibling_s = match info.next_sibling.get() {
                         Some(List { .. }) => "Some(List)",
                         Some(Atom { .. }) => "Some(Atom)",
                         None => "None",
                     };
-                    ds.field("next", &next_s);
+                    ds.field("next_sibling", &next_sibling_s);
                 }
 
                 ds.finish()
@@ -173,12 +170,12 @@ impl<'a> fmt::Debug for Syntax<'a> {
                 if env::var("DFT_VERBOSE").is_ok() {
                     ds.field("highlight", highlight);
                     ds.field("change", &info.change.get());
-                    let next_s = match info.next.get() {
+                    let next_sibling_s = match info.next_sibling.get() {
                         Some(List { .. }) => "Some(List)",
                         Some(Atom { .. }) => "Some(Atom)",
                         None => "None",
                     };
-                    ds.field("next", &next_s);
+                    ds.field("next_sibling", &next_sibling_s);
                 }
 
                 ds.finish()
@@ -250,8 +247,8 @@ impl<'a> Syntax<'a> {
         }
     }
 
-    pub fn next(&self) -> Option<&'a Syntax<'a>> {
-        self.info().next.get()
+    pub fn next_sibling(&self) -> Option<&'a Syntax<'a>> {
+        self.info().next_sibling.get()
     }
 
     pub fn prev_is_contiguous(&self) -> bool {
@@ -399,7 +396,7 @@ fn set_content_id(nodes: &[&Syntax], existing: &mut HashMap<ContentKey, u32>) {
 }
 
 pub fn init_next_prev<'a>(roots: &[&'a Syntax<'a>]) {
-    set_next(roots, None);
+    set_next_sibling(roots);
     set_prev(roots, None);
     set_prev_is_contiguous(roots);
 }
@@ -420,18 +417,13 @@ fn set_unique_id<'a>(nodes: &[&'a Syntax<'a>], next_id: &mut u32) {
     }
 }
 
-/// For every syntax node in the tree, mark the next node according to
-/// a preorder traversal.
-fn set_next<'a>(nodes: &[&'a Syntax<'a>], parent_next: Option<&'a Syntax<'a>>) {
+fn set_next_sibling<'a>(nodes: &[&'a Syntax<'a>]) {
     for (i, node) in nodes.iter().enumerate() {
-        let node_next = match nodes.get(i + 1) {
-            Some(node_next) => Some(*node_next),
-            None => parent_next,
-        };
+        let sibling = nodes.get(i + 1).copied();
+        node.info().next_sibling.set(sibling);
 
-        node.info().next.set(node_next);
         if let List { children, .. } = node {
-            set_next(children, node_next);
+            set_next_sibling(children);
         }
     }
 }
@@ -573,13 +565,34 @@ pub struct MatchedPos {
     pub pos: SingleLineSpan,
 }
 
-// "foo bar" -> vec!["foo", " ", "bar"]
-pub fn split_words(s: &str) -> Vec<String> {
-    lazy_static! {
-        static ref RE: Regex = Regex::new(r"[a-zA-Z0-9]+|\n|[^a-zA-Z0-9\n]").unwrap();
+/// Split `s` into a vec of things that look like words and individual
+/// non-word characters.
+///
+/// "foo bar" -> vec!["foo", " ", "bar"]
+pub fn split_words(s: &str) -> Vec<&str> {
+    let mut res = vec![];
+    let mut word_start = None;
+    for (idx, c) in s.char_indices() {
+        #[allow(clippy::manual_range_contains)]
+        let is_word_constituent =
+            ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || ('0' <= c && c <= '9');
+        if is_word_constituent {
+            if word_start.is_none() {
+                word_start = Some(idx);
+            }
+        } else {
+            if let Some(start) = word_start {
+                res.push(&s[start..idx]);
+                word_start = None;
+            }
+            res.push(&s[idx..idx + 1]);
+        }
     }
 
-    RE.find_iter(s).map(|m| m.as_str().to_owned()).collect()
+    if let Some(start) = word_start {
+        res.push(&s[start..]);
+    }
+    res
 }
 
 fn split_comment_words(
